/*  Botik::Key Pro - Botik Key Client for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2004 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "stdafx.h"
#include "rpc_procs.h"
#include "service.h"
#include "eventmsgs.h"
#include "eventlog.h"

//////////////////////////////////////////////////////////////////////////////
// Global variables.
//////////////////////////////////////////////////////////////////////////////

LPEVENT_LOG_ENTRY	g_lpEventLogHead = NULL;
LPEVENT_LOG_ENTRY	g_lpEventLogTail = NULL;
DWORD				g_dwEventLogCurSize = 0;
CRITICAL_SECTION	g_csEventLogCritSect;
HANDLE				g_hLogFile = NULL;
DWORD				g_dwLogDay = 0;
BOOL				g_fWriteLogFile = FALSE;
BOOL				g_fInformMngr = TRUE;

//////////////////////////////////////////////////////////////////////////////
// Forward declarations.
//////////////////////////////////////////////////////////////////////////////

LPEVENT_LOG_ENTRY EventLogAllocEntry(LONG nMsgTxtLen_);
HANDLE EventLogOpenFile();

//////////////////////////////////////////////////////////////////////////////
// Implementation.
//////////////////////////////////////////////////////////////////////////////

LPEVENT_LOG_ENTRY EventLogAddEntryF(DWORD dwEvtType_, LPCTSTR szFormat_, ...)
{
	LPTSTR            szMsgTxt;
    va_list           pArgs;
	INT               nLen;
	LPEVENT_LOG_ENTRY lpEntry;

    va_start(pArgs, szFormat_);

	nLen = _vsctprintf(szFormat_, pArgs);

	szMsgTxt = new TCHAR[nLen + 1];

	_vstprintf(szMsgTxt, szFormat_, pArgs);

	va_end(pArgs);

	lpEntry = EventLogAddEntry(dwEvtType_, szMsgTxt);

	delete[] szMsgTxt;

	return(lpEntry);
};

LPEVENT_LOG_ENTRY EventLogAddEntry(DWORD dwEvtType_, LPTSTR szMsgTxt_)
{
	SYSTEMTIME			st;

	GetLocalTime(&st);

	EnterCriticalSection(&g_csEventLogCritSect);

	// Log event into the log file.
	if(g_fWriteLogFile)
	{
		LPTSTR szPtr;
		TCHAR  szDateTime[20];
		DWORD  dwSavedSz;

		if(!g_hLogFile || g_dwLogDay != st.wDay) // Open new file 'cause the new day came.
		{
			CloseHandle(g_hLogFile);
			g_hLogFile = EventLogOpenFile();
		};

		szPtr = EventLogTypeToString(dwEvtType_);

		WriteFile(g_hLogFile, szPtr, (DWORD)_tcslen(szPtr), &dwSavedSz, NULL);
		WriteFile(g_hLogFile, ABK_LOGFILE_ENTRY_DELEMITER, 
				  (DWORD)_tcslen(ABK_LOGFILE_ENTRY_DELEMITER), &dwSavedSz, NULL);

		_stprintf(szDateTime, _T("%02d-%02d-%04d %02d:%02d:%02d"), 
				  st.wDay, st.wMonth, st.wYear, st.wHour, st.wMinute, st.wSecond);

		WriteFile(g_hLogFile, szDateTime, (DWORD)_tcslen(szDateTime), &dwSavedSz, NULL);
		WriteFile(g_hLogFile, ABK_LOGFILE_ENTRY_DELEMITER, 
				  (DWORD)_tcslen(ABK_LOGFILE_ENTRY_DELEMITER), &dwSavedSz, NULL);

		WriteFile(g_hLogFile, szMsgTxt_, (DWORD)_tcslen(szMsgTxt_), &dwSavedSz, NULL);
		WriteFile(g_hLogFile, _T("\r\n"), (DWORD)_tcslen(_T("\r\n")), &dwSavedSz, NULL);
	};

	// Put event into the events queue to send them to the manager.
	if(g_fInformMngr)
	{
		LPEVENT_LOG_ENTRY	lpNew;

		lpNew = EventLogAllocEntry((LONG)_tcslen(szMsgTxt_));

		if(!lpNew)
		{
			// XXX Log Event.
		}
		else
		{
			lpNew->dwEventType = dwEvtType_;
			_tcscpy(lpNew->szMsgText, szMsgTxt_);
			lpNew->stTime = st;

			if(g_lpEventLogTail == NULL) // Adding first entry into event log list.
			{
				g_lpEventLogHead = g_lpEventLogTail = lpNew;
			}
			else
			{
				lpNew->lpPrevEntry = g_lpEventLogTail;
				g_lpEventLogTail->lpNextEntry = lpNew;
				g_lpEventLogTail = lpNew;
			};

			// Check if event log list is full. If it full, free first(head) entry.
			if((g_dwEventLogCurSize + 1) > BKP_EVENTLOG_BUFFER_SIZE)
			{
				LPEVENT_LOG_ENTRY lpOldHead;

				lpOldHead = g_lpEventLogHead;
					
				g_lpEventLogHead->lpNextEntry->lpPrevEntry = NULL;
				g_lpEventLogHead = g_lpEventLogHead->lpNextEntry;

				EventLogFreeEntry(lpOldHead);
			}
			else
				g_dwEventLogCurSize++;
		};

		SetEvent(g_hDispMsg02Evt);
	};

	LeaveCriticalSection(&g_csEventLogCritSect);

	return(NULL);
};

LPEVENT_LOG_ENTRY EventLogGetEntry()
{
	LPEVENT_LOG_ENTRY lpEvt = NULL;

	EnterCriticalSection(&g_csEventLogCritSect);

	cfdprintf(_T("EventLogGetEntry: 0x%x, %d"), g_lpEventLogHead, g_dwEventLogCurSize);

	if(g_lpEventLogHead != NULL && g_dwEventLogCurSize > 0)
	{
		lpEvt = g_lpEventLogHead;
		g_lpEventLogHead = lpEvt->lpNextEntry;
		lpEvt->lpNextEntry = NULL;

		if(g_lpEventLogHead != NULL)
			g_lpEventLogHead->lpPrevEntry = NULL;
		else
			g_lpEventLogTail = NULL;

		g_dwEventLogCurSize--;
	};

	LeaveCriticalSection(&g_csEventLogCritSect);

	return(lpEvt);
};

VOID EventLogInit(VOID)
{
	InitializeCriticalSection(&g_csEventLogCritSect);

	if(g_fWriteLogFile)
		g_hLogFile = EventLogOpenFile();
};

HANDLE EventLogOpenFile()
{
	SYSTEMTIME	st;
	HANDLE		hFile;
	TCHAR		szPath[MAX_PATH];
	LPTSTR		ptr;

	// Get module path.
	GetModuleFileName(g_hInst, szPath, MAX_PATH);
	*(ptr = _tcsrchr(szPath, _T('\\'))) = 0;

	// Create directory for log files.
	_tcscat(szPath, _T("\\logfiles"));
	CreateDirectory(szPath, NULL);

	// Create log-file.
	GetLocalTime(&st);
	g_dwLogDay = st.wDay;
	_stprintf(ptr+9, _T("\\%04d-%02d-%02d.log"), st.wYear, st.wMonth, st.wDay);
	cfdprintf(_T("EventLogOpenFile: %s"), szPath);
	hFile = CreateFile(szPath, GENERIC_WRITE|FILE_APPEND_DATA, FILE_SHARE_READ, NULL, 
					   OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hFile == INVALID_HANDLE_VALUE)
	{
		cfInformError(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
					  ABK_MSG_ERR_SVC_OPEN_CONFIG, 
					  _T("Can't open log file."), 
					  GetLastError());
		hFile = NULL;
	};

	SetFilePointer(hFile, 0, NULL, FILE_END); // To append data

	return(hFile);
};

VOID EventLogFree(VOID)
{
	LPEVENT_LOG_ENTRY lpCurEntry;
	
	EnterCriticalSection(&g_csEventLogCritSect);

	lpCurEntry = g_lpEventLogHead;

	g_lpEventLogHead = g_lpEventLogTail = NULL;
	g_dwEventLogCurSize = 0;

	LeaveCriticalSection(&g_csEventLogCritSect);

	while(lpCurEntry != NULL)
	{
		LPEVENT_LOG_ENTRY lpNextEntry;

		lpNextEntry = lpCurEntry->lpNextEntry;

		EventLogFreeEntry(lpCurEntry);

		lpCurEntry = lpNextEntry;
	};
};

LPEVENT_LOG_ENTRY EventLogAllocEntry(LONG nMsgTxtLen_)
{
	LPEVENT_LOG_ENTRY lpNew = NULL;

	lpNew = new EVENT_LOG_ENTRY;

	if(lpNew)
	{
		lpNew->szMsgText = new TCHAR[nMsgTxtLen_ + 1];
		lpNew->lpNextEntry = NULL;
		lpNew->lpPrevEntry = NULL;
		lpNew->dwEventType = 0;
	};

	return(lpNew);
};

VOID EventLogFreeEntry(LPEVENT_LOG_ENTRY lpEntry_)
{
	if(lpEntry_)
	{
		delete[] lpEntry_->szMsgText;
		delete lpEntry_;
	};
};
