/*  Botik::Key Pro - Botik Key Client for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2004 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "service.h"
#include "installation.h"

// Forward declaration...
BOOL InstallNTService(VOID);
BOOL Install9xService(VOID);
BOOL UnInstallNTService(VOID);
BOOL UnInstall9xService(VOID);

INT InstallService(VOID)
{
	BOOL fRet;

	if(g_fIsWinNT)
		fRet = InstallNTService();
	else
		fRet = Install9xService();

	if(fRet)
		cfAlertBox(EVENTLOG_SUCCESS, ABK_SERVICE_NAME, _T("Service installed successfuly."));

	return(fRet ? 0 : 1);
};

INT UnInstallService(VOID)
{
	BOOL fRet;

	if(g_fIsWinNT)
		fRet = UnInstallNTService();
	else
		fRet = UnInstall9xService();

	if(fRet)
		cfAlertBox(EVENTLOG_SUCCESS, ABK_SERVICE_NAME, _T("Service uninstalled successfuly."));

	return(fRet ? 0 : 1);
};

INT StartService()
{
	BOOL fRet = TRUE;

	if(g_fIsWinNT)
	{
		SC_HANDLE  schSCMngr, schSvc;

		cfdprintf("StartService()");

		schSCMngr = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
		if(schSCMngr == NULL)
		{
			cfAlertBoxError(ABK_SERVICE_NAME, 
							_T("Sorry, start service failed, an error occured "
							   "while opening Service Manager."), 
							GetLastError());

			return(FALSE);
		};

		schSvc = OpenService(schSCMngr, ABK_SERVICE_NAME_FOR_SCM, SERVICE_START);
		if(schSvc == NULL) 
		{
			cfAlertBoxError(ABK_SERVICE_NAME, 
							_T("Sorry, start service failed, an error occured "
							   "while opening service."), 
							GetLastError());

			return(FALSE);
		};

		if(!StartService(schSvc, 0, NULL))
		{
			cfAlertBoxError(ABK_SERVICE_NAME, 
							_T("Sorry, start service failed, an error occured "
							   "while starting service."), 
							GetLastError());

			fRet = FALSE;
		};

		CloseServiceHandle(schSvc);
		CloseServiceHandle(schSCMngr);
	}
	else
	{
		TCHAR szPath[MAX_PATH];
		STARTUPINFO si;
		PROCESS_INFORMATION pi;

		GetModuleFileName(g_hInst, szPath, MAX_PATH);

		ZeroMemory(&si, sizeof(STARTUPINFO));
		si.cb = sizeof(si);
		ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));

		if(!CreateProcess(szPath, NULL, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
		{
			cfAlertBoxError(ABK_SERVICE_NAME, 
							_T("Sorry, start service failed, an error occured "
							   "while creating process."), 
							GetLastError());

			fRet = FALSE;
		};

		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	};

	return(fRet ? 0 : 1);
};

BOOL InstallNTService(VOID)
{
	SC_HANDLE			schSCMngr, schSvc;
	SERVICE_DESCRIPTION	sd;
	TCHAR				szPath[MAX_PATH+1];
	HKEY				hKey;
	LONG				lRet;

	schSCMngr = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if(schSCMngr == NULL)
	{
		cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service installation failed, an error occured "
						   "while opening Service Manager."), 
						GetLastError());

		return(FALSE);
	};

	ZeroMemory(szPath, sizeof(szPath));
	GetModuleFileName(g_hInst, szPath, MAX_PATH);
 
    schSvc = CreateService( 
					schSCMngr,						// SCManager database 
					ABK_SERVICE_NAME_FOR_SCM,		// name of service 
					ABK_SERVICE_NAME,				// service name to display 
					SERVICE_ALL_ACCESS,				// desired access 
					SERVICE_WIN32_OWN_PROCESS,		// service type 
					SERVICE_AUTO_START,				// start type 
					SERVICE_ERROR_NORMAL,			// error control type 
					szPath,							// service's binary 
					NULL,							// no load ordering group 
					NULL,							// no tag identifier 
					_T("RPCSS\0\0"), // \0TCPIP\0NTLMSSP
					NULL,							// LocalSystem account 
					NULL);							// no password 
 
    if(schSvc == NULL) 
	{
        cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service installation failed, an error occured "
						   "while creating new service."), 
						GetLastError());

		return(FALSE); 
	};

	sd.lpDescription = _T("Botik::Key Pro Service is alternative client programm for BotikKey system.");

	if(g_fIs2K)
		ChangeServiceConfig2(schSvc, SERVICE_CONFIG_DESCRIPTION, &sd);
 
    CloseServiceHandle(schSvc);
	CloseServiceHandle(schSCMngr);

	// Register event source in the system EventLog.
	lRet = RegCreateKeyEx(HKEY_LOCAL_MACHINE, REG_EVENTLOG_PATH ABK_SERVICE_NAME_FOR_SCM, 
						  0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE|KEY_READ, 
						  NULL, &hKey, NULL);
	if(lRet != ERROR_SUCCESS)
	{
		cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service installation failed, an error occured "
						   "while regustering event source for system EventLog."), 
						lRet);

		return(FALSE);
	};

	cfRegSaveDWVal(hKey, _T("TypesSupported"), 
				   (EVENTLOG_ERROR_TYPE|EVENTLOG_WARNING_TYPE|EVENTLOG_INFORMATION_TYPE));

	RegSetValueEx(hKey, _T("EventMessageFile"), 0, REG_EXPAND_SZ, 
				  (LPBYTE)szPath, (DWORD)_tcslen(szPath)+1);

	RegCloseKey(hKey);


	return(TRUE);
}; // BOOL InstallNTService(VOID)

BOOL Install9xService(VOID)
{
	TCHAR szPath[MAX_PATH+1];
	LONG  lRet;
	HKEY  hKey;

	GetModuleFileName(g_hInst, szPath, MAX_PATH);

	cfdprintf("Path: %s", REG_AUTOSTART_PATH);

	lRet = RegCreateKeyEx(HKEY_LOCAL_MACHINE, REG_AUTOSTART_PATH, 0, NULL, 
						  REG_OPTION_NON_VOLATILE, KEY_WRITE|KEY_READ, 
						  NULL, &hKey, NULL);
	if(lRet != ERROR_SUCCESS)
	{
		cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service installation failed, an error occured "
						   "while creating key in system registry."), 
						GetLastError());

		return(FALSE);
	};

	lRet = cfRegSaveSTRVal(hKey, REG_SVC_AUTOSTART_VAL, szPath);
	if(lRet != ERROR_SUCCESS)
	{
		cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service installation failed, an error occured "
						   "while creating value in system registry."), 
						GetLastError());

		return(FALSE);
	};

	return(TRUE);
}; // BOOL Install9xService(VOID)

BOOL UnInstallNTService(VOID)
{
	SC_HANDLE  schSCMngr, schSvc;
	SERVICE_STATUS ss;

	schSCMngr = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if(schSCMngr == NULL)
	{
		cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service uninstallation failed, an error occured "
						   "while opening Service Manager."), 
						GetLastError());

		return(FALSE);
	};

	schSvc = OpenService(schSCMngr, ABK_SERVICE_NAME_FOR_SCM, DELETE|SERVICE_STOP|SERVICE_QUERY_STATUS);
    if(schSvc == NULL) 
    {
		cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service uninstallation failed, an error occured "
						   "while opening service for delete operation."), 
						GetLastError());

		return(FALSE);
	}; 

	// Query service status to determine the next uninstallation step.
	if(!QueryServiceStatus(schSvc, &ss))
	{
		cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service uninstallation failed, an error occured "
						   "while quereing service status."), 
						GetLastError());

		return(FALSE);
	}
	else
	{
		// If service status is not stopped, so performe stop call.
		// XXX Is this ok to compare to SERVICE_STOPPED only?
		if(ss.dwCurrentState != SERVICE_STOPPED)
		{
			if(!ControlService(schSvc, SERVICE_CONTROL_STOP, &ss))
			{
				cfAlertBoxError(ABK_SERVICE_NAME, 
								_T("Sorry, service uninstallation failed, an error occured "
								   "while stopping service."), 
								GetLastError());

				return(FALSE);
			};
		};
	}; // if(!QueryServiceStatus(schSvc, &ss))
 
    if(!DeleteService(schSvc)) 
    {
		cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service uninstallation failed, an error occured "
						   "while deleting service record from service database."), 
						GetLastError());

		return(FALSE);
	}; 
 
    CloseServiceHandle(schSvc);
	CloseServiceHandle(schSCMngr);

	// Unregister event source in the system EventLog.
	RegDeleteKey(HKEY_LOCAL_MACHINE, REG_EVENTLOG_PATH ABK_SERVICE_NAME_FOR_SCM);

	return(TRUE);
}; // BOOL UnInstallNTService(VOID)

BOOL UnInstall9xService(VOID)
{
	LONG  lRet;
	HKEY  hKey;

	lRet = RegCreateKeyEx(HKEY_LOCAL_MACHINE, REG_AUTOSTART_PATH, 0, NULL, 
						  REG_OPTION_NON_VOLATILE, KEY_WRITE|KEY_READ, 
						  NULL, &hKey, NULL);
	if(lRet != ERROR_SUCCESS)
	{
		cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service uninstallation failed, an error occured "
						   "while opening key in system registry."), 
						GetLastError());

		return(FALSE);
	};

	lRet = RegDeleteValue(hKey, REG_SVC_AUTOSTART_VAL);
	if(lRet != ERROR_SUCCESS)
	{
		cfAlertBoxError(ABK_SERVICE_NAME, 
						_T("Sorry, service installation failed, an error occured "
						   "while deleting value in system registry."), 
						GetLastError());

		return(FALSE);
	};

	return(TRUE);
}; // BOOL UnInstall9xService(VOID)

