/*  Botik::Key Pro - Botik Key Client for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2004 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "service.h"
#include "interface_h.h"
#include "bkeyclient.h"
#include "eventlog.h"
#include "rpc_procs.h"

HANDLE  g_hDispMsg00Evt;
HANDLE  g_hDispMsg01Evt;
HANDLE  g_hDispMsg02Evt;
HANDLE  g_hDispMsg03Evt;

#define NUMBER_OF_EVENTS	4

HANDLE  g_hDispThrd;

BOOL	g_fIsMsgDisp;

#ifdef __cplusplus
extern "C" {
#endif

void SvcMessageDispatcher(void)
{
	DWORD dwEvent;
	BOOL  fLoop;
	unsigned long ulCode;
	HANDLE hHandles[NUMBER_OF_EVENTS];

	cfdprintf("SvcMessageDispatcher");

	// Get dispatcher thread handle.
	if(g_hDispThrd)
		CloseHandle(g_hDispThrd);

	DuplicateHandle(GetCurrentProcess(), GetCurrentThread(), 
					GetCurrentProcess(), &g_hDispThrd, 0, FALSE, 
					DUPLICATE_SAME_ACCESS);

	cfdprintf("SvcMessageDispatcher: Start of SvcMessageDispatcher [%d].", GetCurrentThreadId());

	// Create events for message dispatching.
	hHandles[0] = g_hDispMsg00Evt = CreateEvent(NULL, FALSE, FALSE, NULL);
	hHandles[1] = g_hDispMsg01Evt = CreateEvent(NULL, FALSE, FALSE, NULL);
	hHandles[2] = g_hDispMsg02Evt = CreateEvent(NULL, FALSE, FALSE, NULL);
	hHandles[3] = g_hDispMsg03Evt = CreateEvent(NULL, FALSE, FALSE, NULL);

	g_fIsMsgDisp = TRUE;
	fLoop = TRUE;

	MngrServiceConnected(!ValidateBotikKey(g_szBotikKey));

	while(fLoop)
	{
		cfdprintf(_T("SvcMessageDispatcher: MsgDisp begin WaitForMultipleObjects..."));

		dwEvent = WaitForMultipleObjects(NUMBER_OF_EVENTS, hHandles, FALSE, INFINITE);
		//if(dwEvent != WAIT_OBJECT_0)
		//	continue;

		cfdprintf("SvcMessageDispatcher: MsgDisp send message to the manager: 0x%x", dwEvent);

		RpcTryExcept
		{
			switch(dwEvent)
			{
			case WAIT_OBJECT_0:
				fLoop = FALSE;
				break;

			case WAIT_OBJECT_0+1:
				MngrCurrentAccessMode(g_dwAccessMode, g_dwReqAccessMode);
				break;

			case WAIT_OBJECT_0+2:
				MngrNewEvent();
				break;

			case WAIT_OBJECT_0+3:
				MngrServiceStatus(g_dwCurState | g_dwStatus | g_dwReason);
				break;
			};
		}
		RpcExcept(1)
		{
			ulCode = RpcExceptionCode();
			cfdprintf("SvcMessageDispatcher: Server msg dispatcher exception... 0x%x (%l)", ulCode, ulCode);
			// XXX Log
			fLoop = FALSE;
			g_fIsMsgDisp = FALSE;
		}
		RpcEndExcept
	};

	g_fIsMsgDisp = FALSE;

	// Close event handles.
	CloseHandle(g_hDispMsg00Evt);
	CloseHandle(g_hDispMsg01Evt);
	CloseHandle(g_hDispMsg02Evt);
	CloseHandle(g_hDispMsg03Evt);

	cfdprintf("!SvcMessageDispatcher");
};

int SvcGetServiceStatus()
{
	cfdprintf(_T("SvcGetServiceStatus called. 0x%x"), (g_dwCurState | g_dwStatus));
	return(g_dwCurState | g_dwStatus);
};

int SvcPullEvent(unsigned int* pnEvtType_, LPSYSTEMTIME lpst_, unsigned char** pszMsgTxt_)
{
	LPEVENT_LOG_ENTRY lpEvt;
	INT nRet = 0;

	cfdprintf(_T("SvcPullEvent"));

	lpEvt = EventLogGetEntry();

	*pszMsgTxt_ = NULL;
	*pnEvtType_ = 0;

	if(lpEvt)
	{
		*pszMsgTxt_ = (unsigned char*)midl_user_allocate((_tcslen(lpEvt->szMsgText) + 1) * sizeof(unsigned char));
		
		_tcscpy((char*)*pszMsgTxt_, lpEvt->szMsgText);

		*lpst_ = lpEvt->stTime;

		*pnEvtType_ = lpEvt->dwEventType;

		EventLogFreeEntry(lpEvt);

		nRet = 1;
	};

	cfdprintf(_T("!SvcPullEvent: nRet: %d"), nRet);

	return(nRet);
};

void SvcSetAccessMode(unsigned int nMode_)
{
	cfdprintf(_T("SvcSetAccessMode: Desired access: %d"), nMode_);

	g_dwReqAccessMode = nMode_;

	// Tell BotikKey Client thread that we want to change access mode.
	SetEvent(g_hChangeModeEvt);
};

void SvcGetAccessMode(unsigned int* pnMode_, unsigned int* pnReqMode_)
{
	cfdprintf("SvcGetAccessMode");
	*pnMode_ = g_dwAccessMode;
    *pnReqMode_ = g_dwReqAccessMode;
};

void SvcSetSessionBotikKey(unsigned char* szKey_)
{
	cfdprintf("SvcSetSessionBotikKey: From mngr: %s", "CENSORED ;-)"/*szKey_*/);

	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T("    "));
	
	strcpy(g_szBotikKey, (char*)szKey_);

	// Tell BotikKey Client thread to reconnect, it reconnects with new key.
	SetEvent(g_hChangeModeEvt);
};

void SvcUpdateConfiguration(void)
{
	cfdprintf("SvcUpdateConfiguration");

	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T("    "));

	ServiceLoadConfig();

	if(!ValidateBotikKey(g_szBotikKey))
		MngrPromptPassword();

    // we need it to reconnect to the key server if user change 
    // server address or port number
    SetEvent(g_hChangeModeEvt); 
};

int SvcStopService(void)
{
	INT fRet = TRUE;

	cfdprintf(_T("SvcStopService"));

	if(!g_fIsWinNT)
		ServiceStopping();
	else
	{
		cfdprintf(_T("Unaceptable command, we are running on NT."));
		fRet = FALSE;
	};

	return(fRet);
};

void SvcDisconnectManager(void)
{
	cfdprintf("SvcDisconnectManager");

	SetEvent(g_hDispMsg00Evt);
};

#ifdef __cplusplus
};
#endif

// Memory management.
void __RPC_FAR * __RPC_API midl_user_allocate(size_t nBytes_) 
{ 
    return(malloc(nBytes_)); 
};

void __RPC_API midl_user_free(void __RPC_FAR * ptr_) 
{ 
    free(ptr_); 
};
