/*  Botik::Key Pro - Botik Key Client for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2004 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include <iphlpapi.h>
#include "interface_h.h"
#include "bkeyclient.h"
#include "rpc_procs.h"
#include "eventmsgs.h"
#include "eventlog.h"
#include "service.h"

//////////////////////////////////////////////////////////////////////////////
// Global variables.
//////////////////////////////////////////////////////////////////////////////

SERVICE_STATUS_HANDLE	g_hSvcStat;
SERVICE_STATUS			g_ss;
BOOL					g_fRetCode;
DWORD					g_dwCurState;
DWORD					g_dwReason = 0;

CHAR					g_szBotikKey[ABK_MD5_HASH_LEN+2];
DWORD					g_dwAccessMode;
DWORD					g_dwReqAccessMode;
DWORD					g_dwLastAccessMode;
DWORD					g_dwAccModeWhenDisc;
LPSTR					g_szUserMsg;
DWORD					g_dwUserMsg;
HKEY					g_hKeyLocMch;
DWORD					g_nPasswdPlace;
DWORD					g_dwSvcSwitchStart;
LPSTR					g_szSrvAddr;
USHORT					g_nPortNum;

//////////////////////////////////////////////////////////////////////////////
// Forward declarations.
//////////////////////////////////////////////////////////////////////////////

VOID WINAPI ServiceMainProc(DWORD dwArgc_, LPTSTR* lpszArgv_);
VOID WINAPI ServiceCtrlHandler(DWORD dwCtrl_);

BOOL ServiceInit(VOID);
VOID ServicePerformTask(VOID);
BOOL GetDefaultGateway(LPSTR* szAddr);

//////////////////////////////////////////////////////////////////////////////
// Implementation.
//////////////////////////////////////////////////////////////////////////////

BOOL Service9xStarting(VOID)
{
	g_fRetCode = TRUE;

	ServiceMainProc(0, NULL);

	return(g_fRetCode);
};

BOOL ServiceNTStarting(VOID)
{
	SERVICE_TABLE_ENTRY DispTbl[] = 
	{
		{ABK_SERVICE_NAME_FOR_SCM,	ServiceMainProc},
		{NULL,				        NULL}
	};

	// Log information about starting.
	cfLogEvent(ABK_SERVICE_NAME_FOR_SCM, EVENTLOG_INFORMATION_TYPE, ABK_MSG_INF_STARTING);

	if(!StartServiceCtrlDispatcher(DispTbl))
	{
		cfLogEventError(ABK_SERVICE_NAME_FOR_SCM, ABK_MSG_ERR_START_SVC_CTRL_DISP, 
						_T(""), GetLastError());

		return(FALSE);
	};

	return(TRUE);
};

VOID WINAPI ServiceMainProc(DWORD dwArgc_, LPTSTR* lpszArgv_)
{
	cfdprintf(_T("ServiceMainProc: Starting Botik::Key Pro service [%d]."), GetCurrentThreadId());

	g_hSvcStat = NULL;

	g_dwStatus = ABK_SVC_NOTHING;

	g_ss.dwServiceType      = SERVICE_WIN32_OWN_PROCESS;
	g_ss.dwControlsAccepted = SERVICE_ACCEPT_NETBINDCHANGE|SERVICE_ACCEPT_PARAMCHANGE|
							  SERVICE_ACCEPT_SHUTDOWN|SERVICE_ACCEPT_STOP;
	g_ss.dwWin32ExitCode    = NO_ERROR;
	g_ss.dwServiceSpecificExitCode = 0;
	g_ss.dwCheckPoint       = 0;
	g_ss.dwWaitHint         = 0;

	if(g_fIsWinNT)
	{
		g_hSvcStat = RegisterServiceCtrlHandler(ABK_SERVICE_NAME_FOR_SCM, ServiceCtrlHandler);
		if (g_hSvcStat == (SERVICE_STATUS_HANDLE)0)
		{
			cfLogEventError(ABK_SERVICE_NAME_FOR_SCM, ABK_MSG_ERR_REG_SVC_CTRL_HANDLER, 
						_T(""), GetLastError());

			g_ss.dwCurrentState  = SERVICE_STOPPED;
			g_ss.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
			g_ss.dwServiceSpecificExitCode = ABK_MSG_ERR_REG_SVC_CTRL_HANDLER;
			SetServiceStatus(g_hSvcStat, &g_ss);

			return;
		};
	};

	if(!ServiceInit())
	{
		if(g_fIsWinNT)
		{
			cfLogEvent(ABK_SERVICE_NAME_FOR_SCM, EVENTLOG_ERROR_TYPE, ABK_MSG_ERR_SVC_INIT);

			g_ss.dwCurrentState  = SERVICE_STOPPED;
			g_ss.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
			g_ss.dwServiceSpecificExitCode = ABK_MSG_ERR_SVC_INIT;
			SetServiceStatus(g_hSvcStat, &g_ss);
		};

		g_fRetCode = FALSE;

		return;
	};

	g_dwCurState = SERVICE_RUNNING;
	if(g_fIsWinNT)
	{
		g_ss.dwCurrentState = g_dwCurState;
		SetServiceStatus(g_hSvcStat, &g_ss);

		cfLogEvent(ABK_SERVICE_NAME_FOR_SCM, EVENTLOG_INFORMATION_TYPE, ABK_MSG_INF_RUNNING,
				   _T("\nBotikKey Server: %s\nBotikKey Port: %d"), g_szSrvAddr, g_nPortNum);
	};

	ServicePerformTask();

	// Save state.
	cfRegSaveDWVal(g_hKeyLocMch, REG_LAST_ACCESS_MODE, (g_dwAccessMode & ABK_ACCESS_MASK)); // To save only access mode.

	CloseHandle(g_hNetEvt);
	CloseHandle(g_hChangeModeEvt);
	CloseHandle(g_hShutdownEvt);

	WSACleanup();

	g_dwCurState = SERVICE_STOPPED;
	if(g_fIsWinNT)
	{
		g_ss.dwCurrentState = g_dwCurState;
		SetServiceStatus(g_hSvcStat, &g_ss);

		cfLogEvent(ABK_SERVICE_NAME_FOR_SCM, EVENTLOG_INFORMATION_TYPE, ABK_MSG_INF_STOPPED);
	};

	cfdprintf(_T("ServiceMainProc: Botik::Key Pro service finished."));
};

VOID WINAPI ServiceCtrlHandler(DWORD dwCtrl_)
{
	cfdprintf(_T("ServiceCtrlHandler"));

	switch(dwCtrl_)
	{
	case SERVICE_CONTROL_INTERROGATE:			// System asks for service status.
		g_ss.dwCurrentState = g_dwCurState;
		SetServiceStatus(g_hSvcStat, &g_ss);
		break;

	case SERVICE_CONTROL_PARAMCHANGE:			// System tells to update config.
		EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T("    "));
		ServiceLoadConfig();
		break;

	case SERVICE_CONTROL_STOP:					// System tells to stop server.
	case SERVICE_CONTROL_SHUTDOWN:
		ServiceStopping();
		break;

	case SERVICE_CONTROL_NETBINDADD:			// Network params changed.
	case SERVICE_CONTROL_NETBINDDISABLE:
	case SERVICE_CONTROL_NETBINDENABLE:
	case SERVICE_CONTROL_NETBINDREMOVE:
		cfdprintf(_T("ServiceCtrlHandler: One of NETBIND control fired."));
		break;
	};
};

BOOL ServiceInit(VOID)
{
	LONG	lRet;
	WSADATA	wsaData;

	g_fIsMsgDisp	= FALSE;
	g_dwAccessMode	= ABK_ACCESS_NONE | ABK_ACCESS_CHANGING;
	g_szSrvAddr		= NULL;

	// Opens registry keys to read settings.
	lRet = RegCreateKeyEx(HKEY_LOCAL_MACHINE, REG_SETTINGS_PATH, 0, NULL, 
						  REG_OPTION_NON_VOLATILE, KEY_WRITE|KEY_READ, 
						  NULL, &g_hKeyLocMch, NULL);

	if(lRet != ERROR_SUCCESS)
	{
		cfInformError(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
					  ABK_MSG_ERR_SVC_OPEN_CONFIG, 
					  _T("Service initialization failed. Can't open registry key with configuration."), 
					  lRet);

		return(FALSE);
	};

	// Init WinSock library.
	lRet = WSAStartup(MAKEWORD(2,2), &wsaData);
	if(lRet != 0)
	{
		cfInformError(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
					  ABK_MSG_ERR_WINSOCK_INIT, 
					  _T("Service initialization failed. Can't initialize WinSock library."), 
					  lRet);

		return(FALSE);
	};

	// Init event log.
	EventLogInit();

	cfdprintf("ServiceInit: WinSock: Ver: %d.%d", LOBYTE(wsaData.wVersion), HIBYTE(wsaData.wVersion));
	cfdprintf("ServiceInit: WinSock: Descr: %s", wsaData.szDescription);

	g_hChangeModeEvt = CreateEvent(NULL, FALSE, FALSE, NULL);
	g_hShutdownEvt   = CreateEvent(NULL, FALSE, FALSE, NULL);
	g_hNetEvt        = CreateEvent(NULL, FALSE, FALSE, NULL);

	ZeroMemory(g_szBotikKey, sizeof(g_szBotikKey));

	// Servic is ready to read config and perform its job.
	// Read write log-file param from config to get to know if service must write events into the 
	// log-file. 
	cfRegReadDWVal(g_hKeyLocMch, REG_BKEY_SRV_WRITE_LOG, (LPDWORD)&g_fWriteLogFile, FALSE);
	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T("***"));
	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T(" Botik::Key Pro ..."));

	ServiceLoadConfig();

	return(TRUE);
};

VOID ServicePerformTask(VOID)
{
	// on vista ms depricates sync ncalrpc, now use named pipes
	RPC_STATUS		rpcs;
	unsigned char*	szObjUuid  = (unsigned char*)"9c13db6e-2f5f-434f-8e55-0fbe2816ea5b";
	unsigned char*	szProtSeq;//  = (unsigned char*)"ncacn_np"; //"ncalrpc";     // "ncacn_np";
	unsigned char*	szEndpoint;// = (unsigned char*)"\\pipe\\BotikKeyPro"; // "BotikKeyPro"; // "\\pipe\\BotikKeyPro";

	if(g_fIsWinNT)
	{
		szProtSeq  = (unsigned char*)"ncacn_np"; //"ncalrpc";     // "ncacn_np";
		szEndpoint = (unsigned char*)"\\pipe\\BotikKeyPro"; // "BotikKeyPro"; // "\\pipe\\BotikKeyPro";
	}
	else
	{
		szProtSeq  = (unsigned char*)"ncalrpc";     // "ncacn_np";
		szEndpoint = (unsigned char*)"BotikKeyPro"; // "\\pipe\\BotikKeyPro";
	}

	unsigned char*	szSecurity = (unsigned char*)NULL; 
    unsigned int	cMinCalls  = 1;
    unsigned int	cMaxCalls  = 20;
    unsigned int	fDontWait  = FALSE;

	// Set startup network access mode.
	if(g_dwSvcSwitchStart > ABK_ACCESS_LEAVE)
		g_dwReqAccessMode = g_dwSvcSwitchStart;
	else
		g_dwReqAccessMode = g_dwLastAccessMode;

	g_hDispThrd = NULL;

	// Create and start client thread.
	if(!StartBotikKeyClient())
	{
		cfInformError(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
					  ABK_MSG_ERR_THREAD_EXEC, 
					  _T("Thread execution for BotiKey Client failed."), 
					  GetLastError());

		g_fRetCode = FALSE;

		return;
	};

	// XXX need waiting here, huh?

	// Starting RPC server.
	rpcs = RpcServerRegisterIf(AbkService_v1_0_s_ifspec, NULL, NULL);
	if(rpcs != RPC_S_OK)
	{
		cfInformError(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
					  ABK_MSG_ERR_RPC, 
					  _T("RpcServerRegisterIf failed."), rpcs);

		g_fRetCode = FALSE;

		return;
	};

	rpcs = RpcServerUseProtseqEp(szProtSeq, cMaxCalls, szEndpoint, szSecurity);
	if(rpcs != RPC_S_OK)
	{
		cfInformError(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
					  ABK_MSG_ERR_RPC, 
					  _T("RpcServerUseProtseqEp failed."), rpcs);

		g_fRetCode = FALSE;

		return;
	};

	rpcs = RpcServerListen(cMinCalls, cMaxCalls, fDontWait);
	if(rpcs != RPC_S_OK)
	{
		cfInformError(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
					  ABK_MSG_ERR_RPC, 
					  _T("RpcServerListen failed."), rpcs);

		g_fRetCode = FALSE;

		return;
	};
};

BOOL ServiceLoadConfig(VOID)
{
	BOOL   fUseStdPort;
	BOOL   fNetChangeg = FALSE;
	LPTSTR szPwdPlace = _T(" ");

	cfRegReadDWVal(g_hKeyLocMch, REG_PASSWD_PLACE_VAL, (LPDWORD)&g_nPasswdPlace, 0);
	cfRegReadDWVal(g_hKeyLocMch, REG_SWITCH_WHEN_SVC_START_VAL, (LPDWORD)&g_dwSvcSwitchStart, 0);
	cfRegReadDWVal(g_hKeyLocMch, REG_LAST_ACCESS_MODE, (LPDWORD)&g_dwLastAccessMode, ABK_ACCESS_NONE);
	cfRegReadDWVal(g_hKeyLocMch, REG_BKEY_SRV_PORT_USE_STD, (LPDWORD)&fUseStdPort, 1);
	cfRegReadDWVal(g_hKeyLocMch, REG_BKEY_SRV_WRITE_LOG, (LPDWORD)&g_fWriteLogFile, FALSE);
	cfRegReadDWVal(g_hKeyLocMch, REG_BKEY_SRV_INFORM_MNGR, (LPDWORD)&g_fInformMngr, TRUE);

	if(fUseStdPort)
		g_nPortNum = ABK_BKEY_SRV_DEF_PORT;
	else
		cfRegReadDWVal(g_hKeyLocMch, REG_BKEY_SRV_PORT, (LPDWORD)&g_nPortNum, ABK_BKEY_SRV_DEF_PORT);

	if(!GetServerAddress())
	{
		cfInformError(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
					  ABK_MSG_ERR_BKEY_SRV_ADDR, 
					  _T("Service initialization failed. There is no address of the BotikKey Server, will try to get it late."), 0);
	}
	else
	{
		cfdprintf("ServiceLoadConfig: Address: %s; Port: %d", g_szSrvAddr, g_nPortNum);
	};

	//if(g_nPasswdPlace == ABK_PASSWD_PLACE_SELF)
	//	return(TRUE);

	if(g_nPasswdPlace != ABK_PASSWD_PLACE_SELF) // Load access key.
	{
		ZeroMemory(g_szBotikKey, sizeof(g_szBotikKey));

		if(g_nPasswdPlace == ABK_PASSWD_PLACE_ENV) // Read from environment.
		{
			DWORD dwSize;

			// Get size of specified environment variable if any.
			dwSize = GetEnvironmentVariableA(ABK_BOTIK_KEY_ENV_VAR_NAME, g_szBotikKey, ABK_MD5_HASH_LEN);

			// Check return value.
			if(dwSize == 0 /*|| GetLastError() == ERROR_ENVVAR_NOT_FOUND*/)
			{
				cfInformWarning(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
								ABK_MSG_WRN_NO_KEY_IN_PLACE, 
								_T("Service was configured to take access key from environment "
								"variable called BOTIKKEY.\nBut some troubles occured while getting it."), 
								GetLastError());
			};

			szPwdPlace = _T("  ");
		} 
		else if(g_nPasswdPlace == ABK_PASSWD_PLACE_REG) // Read from registry.
		{
			DWORD dwType, dwSize = ABK_MD5_HASH_LEN+1;
			LONG lRet;

			lRet = RegQueryValueExA(g_hKeyLocMch, ABK_BOTIK_KEY_REG_VAR_NAME, NULL, 
									&dwType, (LPBYTE)g_szBotikKey, &dwSize);

			if(lRet != ERROR_SUCCESS)
			{
				cfInformWarning(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
								ABK_MSG_WRN_NO_KEY_IN_PLACE, 
								_T("Service was configured to take access key from system registry "
								"key.\nBut some troubles occured while getting it."), lRet);
			};

			szPwdPlace = _T("  ");
		} 
		else if(   g_nPasswdPlace == ABK_PASSWD_PLACE_DEF_FILE   // Read from default file.
				|| g_nPasswdPlace == ABK_PASSWD_PLACE_SELF_FILE) // Read from specified file.
		{
			LPTSTR szFile;

			if(g_nPasswdPlace == ABK_PASSWD_PLACE_DEF_FILE)
			{
				TCHAR szModule[MAX_PATH+1];

				GetModuleFileName(g_hInst, szModule, MAX_PATH);

				(*_tcsrchr(szModule, _T('\\'))) = _T(0);

				szFile = new TCHAR[_tcslen(szModule) + _tcslen(ABK_BOTIK_KEY_DEF_FILE_NAME) + 2];
				_tcscpy(szFile, szModule);
				_tcscat(szFile, _T("\\"));
				_tcscat(szFile, ABK_BOTIK_KEY_DEF_FILE_NAME);

				szPwdPlace = _T("  ");
			}
			else
			{
				cfRegReadSTRVal(g_hKeyLocMch, REG_PASSWD_FILE_VAL, &szFile, _T(""));

				szPwdPlace = _T("  ");
			};

			// Open file and read key.
			HANDLE hFile = CreateFile(szFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
			if(hFile == INVALID_HANDLE_VALUE)
			{
				cfInformWarning(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
								ABK_MSG_WRN_NO_KEY_IN_PLACE, 
								_T("Service was configured to take access key from file.\n"
								"But some troubles occured while getting it."), GetLastError());
			}
			else
			{
				DWORD dwReadedSz;
				ReadFile(hFile, g_szBotikKey, ABK_MD5_HASH_LEN, &dwReadedSz, NULL);
				CloseHandle(hFile);
			};

			delete[] szFile;
		};
	};

	//cfdprintf("Key: [%s]", g_szBotikKey);

	// Dump configuration to the log-file. XXX non-Unicode compatible!
	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T("     :"));
	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T(" Key-: %s"), 
					  (g_szSrvAddr[0] ? g_szSrvAddr : _T("  ")));
	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T(" Key-: %d"), g_nPortNum);
	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T("  : %s"), szPwdPlace);
	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T("  %s"), 
					  ValidateBotikKey(g_szBotikKey) ? _T("") : _T("  (  )"));
	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T("   -: %s"), 
					  g_fWriteLogFile ? _T("") : _T(""));
	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T("   : %s"), 
					  g_fInformMngr ? _T("") : _T(""));

	return(TRUE);
};

BOOL GetServerAddress()
{
	BOOL  fUseGWAddr;

	cfRegReadDWVal(g_hKeyLocMch, REG_BKEY_SRV_USE_DEF_GW_ADDR, (LPDWORD)&fUseGWAddr, 1);

	// Get server address.
	if(g_szSrvAddr)
	{
		delete[] g_szSrvAddr;
		g_szSrvAddr = NULL;
	};

	if(fUseGWAddr)
	{
		GetDefaultGateway(&g_szSrvAddr);
	}
	else
	{
		// XXX g_szSrvAddr is not a TCHAR type.
		cfRegReadSTRVal(g_hKeyLocMch, REG_BKEY_SRV_ADDR, &g_szSrvAddr, _T(""));
	};

	if(!g_szSrvAddr[0]) // Validate server address.
	{
		cfdprintf("GetServerAddress: Get server address failed.");

		return(FALSE);
	};

	return(TRUE);
};

BOOL GetDefaultGateway(LPSTR* pszAddr_)
{
	PMIB_IPFORWARDTABLE pIpFwdTbl = NULL;
	DWORD dwSize   = 0;
	DWORD dwStatus = 0;
	BOOL  fRet = FALSE;

	// Get to know how big our buffer must be.
	dwStatus = GetIpForwardTable(pIpFwdTbl, &dwSize, FALSE);
	if (dwStatus == ERROR_INSUFFICIENT_BUFFER) 
	{
		pIpFwdTbl = new MIB_IPFORWARDTABLE[dwSize];

		// Get that table.
		dwStatus = GetIpForwardTable(pIpFwdTbl, &dwSize, FALSE);
	};

	if (dwStatus == ERROR_SUCCESS) 
	{
		// Search for the row in the table we want. The default gateway has a
		// destination of 0.0.0.0.
		for(DWORD i = 0; i < pIpFwdTbl->dwNumEntries; i++) 
		{
			if(pIpFwdTbl->table[i].dwForwardDest == 0) // We've found the default gateway
			{
				LPSTR szAddr = inet_ntoa((*((in_addr*)(LPVOID)&pIpFwdTbl->table[i].dwForwardNextHop)));

				*pszAddr_ = new CHAR[strlen(szAddr) + 1];
				strcpy(*pszAddr_, szAddr);
				fRet = TRUE;
				break;
			};
		};
	};

	if(!fRet)
	{
		*pszAddr_ = new CHAR[1];
		(*pszAddr_)[0] = 0;

		cfInformWarning(g_fIsWinNT, ABK_SERVICE_NAME, ABK_SERVICE_NAME_FOR_SCM, 
						ABK_MSG_WRN_DEF_GATEWAY, 
						_T("Failed to get default gateway address.\n"), dwStatus);

		cfdprintf("GetDefaultGateway: Get default gateway address failed...");
	}
	else
	{
		cfdprintf("GetDefaultGateway: Default gateway address: %s", *pszAddr_);
	};
	
	delete pIpFwdTbl;

	return(fRet);
};

VOID ServiceStopping()
{
	//EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, _T("   "));

	// Tell BotikKey Client thread to stop processing.
	SetEvent(g_hShutdownEvt);

	// Wait for BotikKey client thread finished.
	cfdprintf(_T("ServiceStopping: Waiting for botikkey client thread finished."));
	WaitForSingleObject(g_hClientThrd, INFINITE);
	cfdprintf(_T("ServiceStopping: Botikkey client thread finished."));

	if(g_fIsMsgDisp)
	{
		// Terminate message dispatching loop and wait for dispatching thread finished.
		SetEvent(g_hDispMsg00Evt);
		cfdprintf(_T("ServiceStopping: Waiting for message dispatcher thread finished."));
		// WaitForSingleObject(g_hDispThrd, INFINITE);
		cfdprintf(_T("ServiceStopping: Message dispatcher thread finished."));
	};

	ServiceStopped();
};

VOID ServiceStopped(VOID)
{
	cfdprintf("ServiceStopped");

	// RPC stop.
	RpcMgmtStopServerListening(NULL);
	cfdprintf("ServiceStopped: %d", GetLastError());

	RpcServerUnregisterIf(NULL, NULL, FALSE);
	cfdprintf("ServiceStopped: %d", GetLastError());

	cfdprintf("!ServiceStopped");
};

BOOL ValidateBotikKey(LPSTR szKey_)
{
	INT i = 0;

	while(   szKey_[i] 
		  && (   (szKey_[i] >= '0' && szKey_[i] <= '9') 
			  || (szKey_[i] >= 'a' && szKey_[i] <= 'f') 
			  || (szKey_[i] >= 'A' && szKey_[i] <= 'F')
			 )) 
		i++;

	return(i == ABK_MD5_HASH_LEN ? TRUE : FALSE);
};


